green_dot = "\U0001F7E2"
yellow_dot = "\U0001F7E1"
red_dot = "\U0001F534"
in_progress = "⌛"
checked = "\u2611"
unchecked = "\u2610"
start_bold = "\033[1m"
end_bold = "\033[0m"
bullet = "\u2022"
warning = "⚠️"
data_type_mapping = {
    "string": "String",
    "bigint": "Int64",
    "int": "Int64",
    "smallint": "Int64",
    "boolean": "Boolean",
    "timestamp": "DateTime",
    "date": "DateTime",
    "decimal(38,18)": "Decimal",
    "double": "Double",
}
measure_icon = "\u2211"
table_icon = "\u229E"
column_icon = "\u229F"
